@@ -28,6 +28,7 @@ my $SERVICE_START = 0;
 my $SERVICE_STOP  = 1;
 
 use Init::ServicesList;
+use Utils::Report;
 
 sub get_runlevels
 {
@@ -51,6 +52,7 @@ sub get_runlevels
      "vine-3.0"         => "redhat-6.2",
      "slackware-9.1.0"  => "freebsd-5",
      "gentoo"           => "gentoo",
+     "archlinux"        => "freebsd-5",
      "freebsd-5"        => "freebsd-5",
      "solaris-2.11"     => "freebsd-5",
     );
@@ -58,7 +60,7 @@ sub get_runlevels
   %runlevels=
     (
      "redhat-6.2"      => [ "0", "1", "2", "3", "4", "5", "6" ],
-     "gentoo"          => [ "boot", "default", "nonetwork" ],
+     "gentoo"          => &get_gentoo_runlevels (),
      "freebsd-5"       => [ "default" ],
     );
 
@@ -638,42 +640,55 @@ sub set_bsd_services
 sub get_gentoo_service_status
 {
   my ($script, $runlevel) = @_;
-  my ($services) = &get_gentoo_services_by_runlevel ($runlevel);
+  my ($services) = &get_gentoo_services_for_runlevel ($runlevel);
 
-  foreach $i (@$services)
-  {
-    return 1 if ($i eq $script);
-  }
-
-  return 0;
+  return ($$services {$script});
 }
 
 sub get_gentoo_runlevels
 {
-  my($raw_output) = Utils::File::run_backtick("rc-status -l");
-  my(@runlevels) = split(/\n/,$raw_output);
+  my($raw_output) = Utils::File::run_backtick("rc-status -nc -l");
+  my(@runlevels);
 
-  return @runlevels;
+  return undef if (!$raw_output);
+  @runlevels = split(/\n/,$raw_output);
+
+  return \@runlevels;
 }
 
-sub get_gentoo_services_by_runlevel
+sub get_gentoo_services_for_runlevel
 {
   my($runlevel) = @_;
-  my($raw_output) = Utils::File::run_backtick("rc-status $runlevel");
+  my($raw_output) = Utils::File::run_backtick("rc-status -nc $runlevel");
   my(@raw_lines) = split(/\n/,$raw_output);
-  my(@services);
-  my($line);
+  my($line, $service);
+  my(%services);
 
   foreach $line (@raw_lines)
   {
     if ($line !~ /^Runlevel/)
     {
-      $line=(split(" ",$line))[0];
-	    push(@services,$line);
+      $service = (split(" ",$line))[0];
+      $services{$service} = 1;
 	  }
   }
 
-  return \@services
+  return \%services
+}
+
+sub get_gentoo_runlevels_services
+{
+  my (%runlevels_services, $runlevels);
+
+  $runlevels = &get_gentoo_runlevels ();
+  return undef if (!$runlevels);
+
+  foreach $runlevel (@$runlevels)
+  {
+    $runlevels_services{$runlevel} = &get_gentoo_services_for_runlevel ($runlevel);
+  }
+
+  return \%runlevels_services;
 }
 
 sub get_gentoo_services_list
@@ -705,50 +720,16 @@ sub gentoo_service_exists
   return 0;
 }
 
-sub get_gentoo_runlevels_by_service
+sub get_gentoo_runlevels_status
 {
-  my ($service) = @_;
-  my(@runlevels,@services_in_runlevel,@contain_runlevels, $runlevel);
-  my ($elem);
+  my ($service, $runlevels_services) = @_;
+  my (@arr, $services_in_runlevel);
 
-  # let's do some caching to improve performance
-  if ($gentoo_services_hash eq undef)
+  foreach $runlevel (keys %$runlevels_services)
   {
-    @runlevels = &get_gentoo_runlevels ();
+    $services_in_runlevel = $$runlevels_services {$runlevel};
 
-    foreach $runlevel (@runlevels)
-    {
-      $$gentoo_services_hash{$runlevel} = &get_gentoo_services_by_runlevel ($runlevel);
-    }
-  }
-
-  if (&gentoo_service_exists($service))
-  {
-    foreach $runlevel (keys %$gentoo_services_hash)
-    {
-      $services_in_runlevel = $$gentoo_services_hash {$runlevel};
-
-      foreach $elem (@$services_in_runlevel)
-      {
-        push (@contain_runlevels, $runlevel) if ($elem eq $service);
-      }
-    }
-  }
-
-  return @contain_runlevels;
-}
-
-sub get_gentoo_runlevel_status_by_service
-{
-  my ($service) = @_;
-  my (@arr, @ret);
-  my (@runlevels) = &get_gentoo_runlevels();
-  my (@started) = &get_gentoo_runlevels_by_service($service);
-  my (%start_runlevels) = map { $started[$_], 1 } 0 .. $#started;
-
-  foreach $runlevel (@runlevels)
-  {
-    if (defined $start_runlevels{$runlevel})
+    if ($$services_in_runlevel{$service})
     {
       push @arr, [ $runlevel, $SERVICE_START, 0 ];
     }
@@ -763,26 +744,25 @@ sub get_gentoo_runlevel_status_by_service
 
 sub get_gentoo_service_info
 {
-	my ($service) = @_;
-	my ($script, @actions, @runlevels);
-	my ($role);
-	
-  $runlevels = &get_gentoo_runlevel_status_by_service ($service);
+  my ($service, $runlevels_services) = @_;
+  my (@runlevels_info);
 
-  return ($service, $runlevels);
+  $runlevels_info = &get_gentoo_runlevels_status ($service, $runlevels_services);
+
+  return ($service, $runlevels_info);
 }
 
 sub get_gentoo_services
 {
-  my ($service);
-  my (@arr);
+  my ($service, @arr);
   my ($service_list) = &get_gentoo_services_list ();
+  my ($runlevels_services) = &get_gentoo_runlevels_services ();
 
   foreach $service (@$service_list)
   {
     my (@info);
 
-    @info = &get_gentoo_service_info ($service);
+    @info = &get_gentoo_service_info ($service, $runlevels_services);
     push @arr, \@info if scalar (@info);
   }
 
@@ -813,7 +793,14 @@ sub run_gentoo_script
 
 sub set_gentoo_service_status
 {
-  my ($script, $rl, $action) = @_;
+  my ($script, $rl, $status, $runlevels_services) = @_;
+  my ($services_in_runlevel, $old_status);
+
+  $services_in_runlevel = $$runlevels_services {$rl};
+  $old_status = ($$services_in_runlevel{$script}) ?
+      $SERVICE_START : $SERVICE_STOP;
+
+  return if ($status == $old_status);
 
   if ($action == $SERVICE_START)
   {
@@ -832,6 +819,9 @@ sub set_gentoo_services
 {
   my ($services) = @_;
   my ($action, $rl, $script, $arr);
+  my ($runlevels_services) = &get_gentoo_runlevels_services ();
+
+  return if (!$runlevels_services);
 
   foreach $service (@$services)
   {
@@ -842,7 +832,8 @@ sub set_gentoo_services
     {
       $action = $$i[1];
       $rl = $$i[0];
-      &set_gentoo_service_status ($script, $rl, $action);
+      &set_gentoo_service_status ($script, $rl, $action,
+                                  $runlevels_services);
     }
   }
 }
@@ -856,7 +847,7 @@ sub get_rcng_status_by_service
   # This is the only difference between rcNG and archlinux
   if ($Utils::Backend::tool{"platform"} eq "archlinux")
   {
-      return &Utils::File::exists ("/var/run/daemons/$service");
+    return &Utils::File::exists ("/var/run/daemons/$service");
   }
   else
   {
@@ -892,7 +883,7 @@ sub get_rcng_service_info
     push @runlevels, [ "default", $SERVICE_STOP, 0 ];
   }
 
-  return ($script, $runlevels);
+  return ($script, \@runlevels);
 }
 
 sub get_rcng_services
@@ -920,13 +911,13 @@ sub run_rcng_script
 
   if (!&Utils::File::run ("/etc/rc.d/$service $arg"))
   {
-    &Utils::File::do_report ("service_sysv_op_success", $service, $arg);
-    &Utils::File::leave ();
+    &Utils::Report::do_report ("service_sysv_op_success", $service, $arg);
+    &Utils::Report::leave ();
     return 0;
   }
 
-  &Utils::File::do_report ("service_sysv_op_failed", $service, $arg);
-  &Utils::File::leave ();
+  &Utils::Report::do_report ("service_sysv_op_failed", $service, $arg);
+  &Utils::Report::leave ();
   return -1;
 }
 
@@ -987,8 +978,8 @@ sub set_archlinux_service_status
   my ($script, $active) = @_;
   my $rcconf = '/etc/rc.conf';
   my ($daemons);
-
   $daemons = &Utils::Parse::get_sh ($rcconf, "DAEMONS");
+  $daemons =~ s/[()]//;
 
   if (($daemons =~ m/$script/) && !$active)
   {
@@ -999,7 +990,7 @@ sub set_archlinux_service_status
     $daemons =~ s/network/network $script/g;
   }
 
-  &Utils::Replace::set_sh ($rcconf, "DAEMONS", $daemons);
+  &Utils::Replace::set_sh ($rcconf, "DAEMONS", "($daemons)", 1);
   &run_rcng_script ($service, ($active) ? "start" : "stop");
 }
 
@@ -1288,7 +1279,7 @@ sub get_init_type
   {
     return "rcng";
   }
-  elsif (($gst_dist =~ /gentoo/) || ($gst_dist =~ /^vlos/))
+  elsif ($gst_dist =~ /gentoo/)
   {
     return "gentoo";
   }
